from yasmon.processor import YAMLProcessor
from yasmon.loggers import LoggerSyntaxError
from yasmon.loggers import JournalLogger
from yasmon.loggers import StdErrLogger
from yasmon.loggers import FileLogger

import unittest


class JournalLoggerTest(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(JournalLoggerTest, self).__init__(*args, **kwargs)
        self.proc = YAMLProcessor()

    def test_from_yaml_raise_exceptions(self):
        """
        Test if JournalLogger.from_yaml() raises appropriate exceptions.
        """

        data = """
        level: ]]
        """
        self.assertRaises(LoggerSyntaxError, JournalLogger.from_yaml, data)


class StdErrLoggerTest(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(StdErrLoggerTest, self).__init__(*args, **kwargs)
        self.proc = YAMLProcessor()

    def test_from_yaml_raise_exceptions(self):
        """
        Test if StdErrLogger.from_yaml() raises appropriate exceptions.
        """

        data = """
        level: ]]
        """
        self.assertRaises(LoggerSyntaxError, StdErrLogger.from_yaml, data)


class FileLoggerTest(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(FileLoggerTest, self).__init__(*args, **kwargs)
        self.proc = YAMLProcessor()

    def test_from_yaml_raise_exceptions(self):
        """
        Test if FileLogger.from_yaml() raises appropriate exceptions.
        """

        data = """
        level: ]]
        """
        self.assertRaises(LoggerSyntaxError, FileLogger.from_yaml, data)

        data = """
        level: info
        path:
            - /invalid/syntax
        """
        self.assertRaises(LoggerSyntaxError, FileLogger.from_yaml, data)


if __name__ == '__main__':
    unittest.main(verbosity=2)



from yasmon.processor import YAMLProcessor
from yasmon.tasks import WatchfilesTask, TaskSyntaxError
from yasmon.tasks import TaskError, TaskRunner, TaskList
from yasmon.callbacks import CallbackAttributeError
from yasmon.callbacks import CallbackCircularAttributeError

import watchfiles
import unittest
import time
import subprocess


class TaskRunnerTest(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(TaskRunnerTest, self).__init__(*args, **kwargs)
        self.proc = YAMLProcessor()
        self.input_producer_script = 'tests/assets/watchfiles_test.sh'
        self.input_producer = None

    def start_input_producer(self):
        self.input_producer = subprocess.Popen(self.input_producer_script,
                                               stdout=subprocess.PIPE,
                                               stderr=subprocess.PIPE)

        # wait a moment for input producer (just to be safe)
        time.sleep(1)

    def stop_input_producer(self):
        if self.input_producer:
            self.input_producer.kill()
            self.input_producer.wait()
            self.input_producer.communicate()
            self.input_producer = None

    def test_call_testenv(self):
        """
        Test TaskRunner.__call__() tasks cancel if testenv == True.
        """

        test_yaml = """
        log_stderr:
            level: error
        callbacks:
            callback0:
                type: shell
                command: exit 0;
        tasks:
            watchfilestask:
                type: watchfiles
                changes:
                    - deleted
                paths:
                    - tests/assets/tmp/
                callbacks:
                    - callback0
                attrs:
                    myattr: somevalue
        """
        self.start_input_producer()
        self.proc.load_document(test_yaml)
        callbacks = self.proc.get_callbacks()
        tasks = self.proc.get_tasks(callbacks)
        runner = TaskRunner(tasks, testenv=True)
        self.stop_input_producer()

        try:
            runner.loop.run_until_complete(runner())
            exited = True
        except Exception:
            raise

        assert exited is True

    def test_call_propagate_exceptions(self):
        """
        Test that TaskRunner.__call__() propagates exceptions.
        """

        # TaskError
        test_yaml = """
        callbacks:
            callback0:
                type: shell
                command: exit 0
        tasks:
            watchfilestask:
                type: watchfiles
                changes:
                    - added
                    - modified
                    - deleted
                paths:
                    - file/not/found
                callbacks:
                    - callback0
                max_retry: 1
                timeout: 1
        """
        self.proc.load_document(test_yaml)
        callbacks = self.proc.get_callbacks()
        tasks = self.proc.get_tasks(callbacks)
        runner = TaskRunner(tasks)
        fun = runner.loop.run_until_complete
        self.assertRaises(TaskError, fun, runner())

        # CallbackCircularAttributeError
        test_yaml = """
        callbacks:
            callback0:
                type: shell
                command: exit 0; {WRONG}
        tasks:
            watchfilestask:
                type: watchfiles
                changes:
                    - modified
                paths:
                    - tests/assets/tmp/watchfiles_call_test
                callbacks:
                    - callback0
                attrs:
                    myattr: somevalue using {WRONG}
                    WRONG: "{myattr}"
        """
        self.start_input_producer()
        self.proc.load_document(test_yaml)
        callbacks = self.proc.get_callbacks()
        tasks = self.proc.get_tasks(callbacks)
        runner = TaskRunner(tasks, testenv=True)
        fun = runner.loop.run_until_complete
        self.assertRaises(CallbackCircularAttributeError, fun, runner())
        self.stop_input_producer()

        # CallbackAttributesError
        test_yaml = """
        callbacks:
            callback0:
                type: shell
                command: exit 0; {WRONG}
        tasks:
            watchfilestask:
                type: watchfiles
                changes:
                    - added
                paths:
                    - tests/assets/tmp/
                callbacks:
                    - callback0
                attrs:
                    myattr: somevalue
        """
        self.start_input_producer()
        self.proc.load_document(test_yaml)
        callbacks = self.proc.get_callbacks()
        tasks = self.proc.get_tasks(callbacks)
        runner = TaskRunner(tasks, testenv=True)
        fun = runner.loop.run_until_complete
        self.assertRaises(CallbackAttributeError, fun, runner())
        self.stop_input_producer()

        # TaskError
        test_yaml = """
        callbacks:
            callback0:
                type: shell
                command: exit 0;
        tasks:
            watchfilestask:
                type: watchfiles
                changes:
                    - added
                    - deleted
                    - modified
                paths:
                    - tests/assets/tmp/DOES_NOT_EXIST
                callbacks:
                    - callback0
                max_retry: 1
                timeout: 1
        """
        self.proc.load_document(test_yaml)
        callbacks = self.proc.get_callbacks()
        tasks = self.proc.get_tasks(callbacks)
        runner = TaskRunner(tasks)
        fun = runner.loop.run_until_complete
        self.assertRaises(TaskError, fun, runner())


class WatchfilesTaskTest(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(WatchfilesTaskTest, self).__init__(*args, **kwargs)
        self.proc = YAMLProcessor()
        self.input_producer_script = 'tests/assets/watchfiles_test.sh'
        self.input_producer = None

    def start_input_producer(self):
        self.input_producer = subprocess.Popen(self.input_producer_script,
                                               stdout=subprocess.PIPE,
                                               stderr=subprocess.PIPE)

        # wait a moment for input producer (just to be safe)
        time.sleep(1)

    def stop_input_producer(self):
        if self.input_producer:
            self.input_producer.kill()
            self.input_producer.wait()
            self.input_producer.communicate()
            self.input_producer = None

    def test_from_yaml_value_assignment(self):
        """
        Test WatchfilesTask.from_yaml() for assignment of values.
        """

        # default values
        data = """
        changes:
            - added
            - modified
            - deleted
        paths:
            - tests/assets/config.yaml
            - tests/assets/invalid.yaml
        """
        task = WatchfilesTask.from_yaml("name", data, [])
        assert task.name == 'name'
        assert task.paths == [
            'tests/assets/config.yaml',
            'tests/assets/invalid.yaml'
        ]
        assert task.changes == [
            watchfiles.Change.added,
            watchfiles.Change.modified,
            watchfiles.Change.deleted
        ]
        assert task.attrs == {}
        assert task.timeout == 30
        assert task.max_retry == 5

        # fully specified
        data = """
        changes:
            - added
            - modified
            - deleted
        paths:
            - tests/assets/config.yaml
            - tests/assets/invalid.yaml
        attrs:
            myattr1: value1
            myattr2: value2
        timeout: 10
        max_retry: 7
        """
        task = WatchfilesTask.from_yaml("name", data, [])
        assert task.name == 'name'
        assert task.paths == [
            'tests/assets/config.yaml',
            'tests/assets/invalid.yaml'
        ]
        assert task.changes == [
            watchfiles.Change.added,
            watchfiles.Change.modified,
            watchfiles.Change.deleted
        ]
        assert task.attrs == {
            'myattr1': 'value1',
            'myattr2': 'value2',
        }
        assert task.timeout == 10
        assert task.max_retry == 7

    def test_from_yaml_raise_TaskSyntaxError(self):
        """
        Test if WatchfilesTask.from_yaml() raises TaskSyntaxError
        """
        fun = WatchfilesTask.from_yaml

        # general yaml syntax error
        data = """
        changes:
            - added
            - modified
            - deleted
        paths: ][
        attrs:
            myattr1: value1
            myattr2: value2
        """
        self.assertRaises(TaskSyntaxError, fun, "name", data, [])

        # missing changes dictionary
        data = """
        paths:
            - some/path
        attrs:
            myattr1: value1
            myattr2: value2
        """
        self.assertRaises(TaskSyntaxError, fun, "name", data, [])

        # missing paths dictionary
        data = """
        changes:
            - added
            - modified
            - deleted
        attrs:
            myattr1: value1
            myattr2: value2
        """
        self.assertRaises(TaskSyntaxError, fun, "name", data, [])

        # empty paths dictionary
        data = """
        changes:
            - added
            - modified
            - deleted
        paths:
        attrs:
            myattr1: value1
            myattr2: value2
        """
        self.assertRaises(TaskSyntaxError, fun, "name", data, [])

        # empty paths list
        data = """
        changes:
            - added
            - modified
            - deleted
        paths: []
        attrs:
            myattr1: value1
            myattr2: value2
        """
        self.assertRaises(TaskSyntaxError, fun, "name", data, [])

        # paths is not a list
        data = """
        changes:
            - added
            - modified
            - deleted
        paths:
            /some/path1:
            /some/path2:
        attrs:
            myattr1: value1
            myattr2: value2
        """
        self.assertRaises(TaskSyntaxError, fun, "name", data, [])

        # paths is not a string
        data = """
        changes:
            - added
            - modified
            - deleted
        paths:
            - []
        attrs:
            myattr1: value1
            myattr2: value2
        """
        self.assertRaises(TaskSyntaxError, fun, "name", data, [])

        # changes is not a list
        data = """
        changes:
            added:
            modified:
            deleted:
        paths:
            - tests/assets/config.yaml
            - tests/assets/invalid.yaml
        attrs:
            myattr1: value1
            myattr2: value2
        """
        self.assertRaises(TaskSyntaxError, fun, "name", data, [])

        # attrs not a dictionary
        data = """
        changes:
            - added
            - modified
            - deleted
        paths:
            - tests/assets/config.yaml
            - tests/assets/invalid.yaml
        attrs:
            - myattr1: value1
            - myattr2: value2
        """
        self.assertRaises(TaskSyntaxError, fun, "name", data, [])

        # changes values invalid
        data = """
        changes:
            - added
            - INVALID
            - deleted
        paths:
            - tests/assets/config.yaml
            - tests/assets/invalid.yaml
        attrs:
            myattr1: value1
            myattr2: value2
        """
        self.assertRaises(TaskSyntaxError, fun, "name", data, [])

        # invalid max_retry (due to ValueError)
        data = """
        changes:
            - added
            - deleted
        paths:
            - tests/assets/config.yaml
            - tests/assets/invalid.yaml
        attrs:
            myattr1: value1
            myattr2: value2
        max_retry: INVALID
        """
        self.assertRaises(TaskSyntaxError, fun, "name", data, [])

        # invalid max_retry (due to TypeError)
        data = """
        changes:
            - added
            - deleted
        paths:
            - tests/assets/config.yaml
            - tests/assets/invalid.yaml
        attrs:
            myattr1: value1
            myattr2: value2
        max_retry:
            INVALID:
        """
        self.assertRaises(TaskSyntaxError, fun, "name", data, [])

        # invalid timeout (due to ValueError)
        data = """
        changes:
            - added
            - deleted
        paths:
            - /some/path1
            - /some/path2
        attrs:
            myattr1: value1
            myattr2: value2
        timeout: INVALID
        """
        self.assertRaises(TaskSyntaxError, fun, "name", data, [])

        # invalid timeout (due to TypeError)
        data = """
        changes:
            - added
            - deleted
        paths:
            - tests/assets/config.yaml
            - tests/assets/invalid.yaml
        attrs:
            myattr1: value1
            myattr2: value2
        timeout:
            - INVALID
        """
        self.assertRaises(TaskSyntaxError, fun, "name", data, [])

        # invalid timeout (must be positive)
        data = """
        changes:
            - added
            - deleted
        paths:
            - tests/assets/config.yaml
            - tests/assets/invalid.yaml
        attrs:
            myattr1: value1
            myattr2: value2
        timeout: 0
        """
        self.assertRaises(TaskSyntaxError, fun, "name", data, [])

    def test_call_testenv(self):
        """
        Test if WatchfilesTask.__call__() runs through.
        """

        # TaskError on watched file deleted
        test_yaml = """
        callbacks:
            callback0:
                type: shell
                command: exit 0;
        tasks:
            watchfilestask:
                type: watchfiles
                changes:
                    - modified
                    - deleted
                    - added
                paths:
                    - tests/assets/tmp/
                callbacks:
                    - callback0
                max_retry: 2
                timeout: 1
        """
        self.start_input_producer()
        self.proc.load_document(test_yaml)
        callbacks = self.proc.get_callbacks()
        tasks = self.proc.get_tasks(callbacks)
        runner = TaskRunner(tasks, testenv=True)
        fun = runner.loop.run_until_complete
        try:
            fun(runner())
        except Exception:
            self.fail()
        self.stop_input_producer()

    def test_call_raise_exceptions(self):
        """
        Test if WatchfilesTask.__call__() raises TaskError.
        """

        # TaskError on watched file deleted
        test_yaml = """
        callbacks:
            callback0:
                type: shell
                command: exit 0;
        tasks:
            watchfilestask:
                type: watchfiles
                changes:
                    - modified
                paths:
                    - tests/assets/tmp/watchfiles_test.sh.pid
                callbacks:
                    - callback0
                max_retry: 2
                timeout: 1
        """
        self.start_input_producer()
        self.proc.load_document(test_yaml)
        callbacks = self.proc.get_callbacks()
        tasks = self.proc.get_tasks(callbacks)
        runner = TaskRunner(tasks)
        fun = runner.loop.run_until_complete
        self.assertRaises(TaskError, fun, runner())
        self.stop_input_producer()


class TaskListTest(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(TaskListTest, self).__init__(*args, **kwargs)

    def test_construction(self):
        """
        Test TaskList.__init__().
        """

        data = """
        changes:
            - added
        paths:
            - /
        """
        task = WatchfilesTask.from_yaml("name", data, [])

        # empty task list
        tasklist = TaskList()
        assert len(tasklist) == 0

        # taskl list form an iterable
        tasklist = TaskList([task, task, task])
        assert len(tasklist) == 3


if __name__ == '__main__':
    unittest.main(verbosity=2)



from loguru import logger
from abc import ABC, abstractmethod
from typing import Self, TYPE_CHECKING
import asyncio
import yaml
import re
import smtplib
import email
import ssl
import mimetypes
import pathlib

if TYPE_CHECKING:
    from .tasks import AbstractTask


def process_attributes(expr: str, attrs: dict[str, str]) -> str:
    """
    Performs attribute substitutions in `expr`.

    This is also checking for circular dependencies (`max_depth = 42`).

    :param expr: expression to process
    :param attrs: attributes to substitute

    :raises CallbackCircularAttributeError: see documentation
    :raises CallbackAttributeError: see documentation

    :return: processed expression
    :rtype: str
    """
    max_depth = 42
    regex = re.compile(r"\{[^{}]*\}")  # match attributes
    search = regex.search(expr)

    depth = 0
    while search:
        try:
            expr = expr.format(**attrs)
        except KeyError as e:
            raise CallbackAttributeError(e)

        if depth < max_depth:
            depth += 1
        else:
            raise CallbackCircularAttributeError(expr)

        search = regex.search(expr)

    return expr


class CallbackNotImplementedError(Exception):
    """
    Raised if a requested callback type is not implemented.
    """

    def __init__(self, type: str, message="callback {type} not implemented"):
        self.message = message.format(type=type)
        super().__init__(self.message)


class CallbackAttributeError(Exception):
    """
    Raised when an undefined task attribute is used by a callback.
    """

    def __init__(self, attr: str, message="undefined attribute {attr}"):
        self.message = message.format(attr=attr)
        super().__init__(self.message)


class CallbackCircularAttributeError(Exception):
    """
    Raised when task attributes have circular dependencies.
    """

    def __init__(self, expr: str,
                 message="{expr}\ndetected circular attributes"):
        self.message = message.format(expr=expr)
        super().__init__(self.message)


class CallbackSyntaxError(Exception):
    """
    Raised on callback syntax error.
    """

    def __init__(self, message="callback syntax error"):
        self.message = message
        super().__init__(self.message)


class CallbackError(Exception):
    """
    Raised on callback execution error.
    """

    def __init__(self, message="callback error"):
        self.message = message
        super().__init__(self.message)


class CallbackDict(dict):
    """
    A dedicated `dictionary` for callbacks.
    """

    def __init__(self, mapping=(), **kwargs):
        super().__init__(mapping, **kwargs)


class AbstractCallback(ABC):
    """
    Abstract class from which all callback classes are derived.

    Derived callbacks are functors and can be used as coroutines for any
    of :class:`yasmon.tasks.AbstractTasks`.

    The preferred way to instatiate a callback is from class
    method :func:`~from_yaml`.
    """

    @abstractmethod
    def __init__(self):
        if not self.name:
            self.name = "Generic Callback"
        logger.info(f'{self.name} ({self.__class__}) initialized')

    @abstractmethod
    async def __call__(self, task: 'AbstractTask', attrs: dict[str, str]):
        """
        Coroutine called by :class:`TaskRunner`.

        :param task: task calling the callback
        :raises CallbackError: see documentation
        :raises CallbackAttributeError: see documentation
        :raises CallbackCircularAttributeError: see documentation
        """
        logger.info(f'{self.name} ({self.__class__}) called by '
                    f'{task.name} ({task.__class__})')

    @classmethod
    @abstractmethod
    def from_yaml(cls, name: str, data: str):
        """
        A class method for constructing a callback from a YAML document.

        :param name: unique identifier
        :param data: yaml data

        :return: new instance
        """
        logger.debug(f'{name} defined form yaml \n{data}')


class ShellCallback(AbstractCallback):
    """
    Callback implementing shell command execution.
    """

    def __init__(self, name: str, cmd: str) -> None:
        """
        :param name: unique identifier
        :param cmd: command to be executed
        """
        self.name = name
        self.cmd = cmd
        super().__init__()

    async def __call__(self, task: 'AbstractTask', attrs: dict[str, str]):
        await super().__call__(task, attrs)

        try:
            cmd = process_attributes(self.cmd, attrs)
        except CallbackAttributeError:
            raise
        except CallbackCircularAttributeError:
            raise

        proc = await asyncio.create_subprocess_shell(
            cmd,
            stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE)

        stdout, stderr = await proc.communicate()
        out = stdout.decode()
        err = stderr.decode()

        if out:
            logger.info(f'callback {self.name} stdout:\n {out}')

        if err:
            logger.error(f'callback {self.name} stderr:\n {err}')

        return stdout, stderr

    @classmethod
    def from_yaml(cls, name: str, data: str) -> Self:
        """
        :class:`ShellCallback` can be also constructed from a YAML snippet.

        .. code:: yaml

            command: ls -lah /path/to/some/dir/

        :param name: unique identifier
        :param data: YAML snippet

        :return: new instance
        :rtype: ShellCallback
        """
        super().from_yaml(name, data)
        try:
            parsed = yaml.safe_load(data)
        except yaml.YAMLError as err:
            raise CallbackSyntaxError(err)

        if 'command' not in parsed:
            raise CallbackSyntaxError(f"""\
            in callback {name} missig command
            """)

        if not isinstance(parsed['command'], str):
            raise CallbackSyntaxError(f"""\
            in callback {name} not a string
            """)

        cmd = parsed["command"]
        return cls(name, cmd)


class LoggerCallback(AbstractCallback):
    """
    Callback implementing logger calls
    """

    def __init__(self, name: str, level: str, message: str) -> None:
        """
        :param name: unique identifier
        :param level: logging level
        :param message: message to pass to logger
        """
        self.name = name
        self.level = level
        self.message = message
        super().__init__()

    async def __call__(self, task: 'AbstractTask',
                       attrs: dict[str, str]) -> None:
        await super().__call__(task, attrs)

        try:
            message = process_attributes(self.message, attrs)
        except CallbackAttributeError:
            raise
        except CallbackCircularAttributeError:
            raise

        method = getattr(logger, self.level)
        method(message)

    @classmethod
    def from_yaml(cls, name: str, data: str) -> Self:
        """
        :class:`LoggerCallback` can be also constructed from a YAML snippet.

        .. code:: yaml

            level: [error | info | debug | ... (see Loguru docs)]
            message: message

        :param name: unique identifier
        :param data: YAML snippet

        :return: new instance
        :rtype: LoggerCallback
        """
        super().from_yaml(name, data)
        try:
            parsed = yaml.safe_load(data)
        except yaml.YAMLError as err:
            raise CallbackSyntaxError(err)

        imp_levels = [
            'trace',
            'debug',
            'info',
            'success',
            'warning',
            'error',
            'critical'
        ]

        if 'level' not in parsed:
            raise CallbackSyntaxError(f"""\
            in callback {name} missig logger level
            """)

        level = parsed["level"]
        if level not in imp_levels:
            raise CallbackSyntaxError(f"""\
            in callback {name} invalid logger level {level}
            """)

        if 'message' not in parsed:
            raise CallbackSyntaxError(f"""\
            in callback {name} missig message
            """)

        if not isinstance(parsed['message'], str):
            raise CallbackSyntaxError(f"""\
            in callback {name} message not a string
            """)

        message = parsed["message"]
        return cls(name, level, message)


class MailCallback(AbstractCallback):
    """
    Callback implementing sending simple notification mails with
    attachments.
    """

    def __init__(self, name: str, host: str, port: int, login: str,
                 password: str, security: str, toaddr: str, subject: str,
                 fromaddr: str, message: str, attach: list[str],
                 delay: int) -> None:
        """
        :param name: unique callback identifier
        :param host: smtp host
        :param port: smtp port
        :param login: login name
        :param password: password
        :param security: ``starttls`` or ``ssl``
        :param toaddr: to address
        :param subject: mail subject
        :param fromaddr: from address
        :param message: message
        """
        self.name = name
        self.host = host
        self.port = port
        self.login = login
        self.password = password
        self.security = security
        self.toaddr = toaddr
        self.subject = subject
        self.fromaddr = fromaddr
        self.message = message
        self.attach = attach
        self.delay = delay
        super().__init__()

    def send_message_starttls(self, message: email.message.EmailMessage):
        """
        Send message using STARTTLS.
        """

        context = ssl.create_default_context()
        server = smtplib.SMTP(self.host, self.port)
        try:
            server.starttls(context=context)
            server.login(self.login, self.password)
            server.send_message(message)
        except Exception:
            raise
        finally:
            server.quit()

    def send_message_ssl(self, message: email.message.EmailMessage):
        """
        Send message using SSL.
        """

        try:
            server = smtplib.SMTP_SSL(self.host, self.port)
            server.login(self.login, self.password)
            server.send_message(message)
        except Exception:
            raise
        finally:
            server.quit()

    def process_attachments(self, message: email.message.EmailMessage,
                            attrs: dict[str, str]) -> None:
        """
        Process list of attachments from ``attach`` and attach these to
        ``message``.

        :raises CallbackError: if path to file does not exist
        """

        try:
            for attachment in self.attach:
                processed_path = process_attributes(attachment, attrs)
                filepath = pathlib.Path(processed_path)
                filename = filepath.name
                mimetype = mimetypes.guess_type(filename)
                if mimetype[0]:
                    data = mimetype[0].split('/')
                    attachment_mimetype = {
                        'maintype': data[0],
                        'subtype': data[1]
                    }
                else:
                    attachment_mimetype = {
                        'maintype': 'text',
                        'subtype': 'plain'
                    }
                with open(filepath, 'rb') as fh:
                    attachment_content = fh.read()
                    message.add_attachment(
                        attachment_content,
                        maintype=attachment_mimetype.get('maintype'),
                        subtype=attachment_mimetype.get('subtype'),
                        filename=filename)
        except Exception as err:
            raise CallbackError(
                f"in callback '{self.name}' exception "
                f"'{err.__class__.__name__}' was raised, {err}")

    async def __call__(self, task: 'AbstractTask',
                       attrs: dict[str, str]) -> None:
        await super().__call__(task, attrs)

        if self.delay > 0:
            logger.debug(f'{self.name} ({self.__class__}) '
                         f'delayed for {self.delay}')
            await asyncio.sleep(self.delay)

        try:
            message = email.message.EmailMessage()
            message['Subject'] = process_attributes(self.subject, attrs)
            message['From'] = process_attributes(self.fromaddr, attrs)
            message['To'] = process_attributes(self.toaddr, attrs)
            content = process_attributes(self.message, attrs)
            message.set_content(content)
            self.process_attachments(message, attrs)
        except CallbackError:
            raise
        except CallbackAttributeError:
            raise
        except CallbackCircularAttributeError:
            raise

        try:
            loop = asyncio.get_event_loop()
            match self.security:
                case 'starttls':
                    await loop.run_in_executor(
                        None, self.send_message_starttls, message)
                case 'ssl':
                    await loop.run_in_executor(
                        None, self.send_message_ssl, message)
            logger.debug(f'{self.name} ({self.__class__}) '
                         f'sent a mail')
        except Exception as err:
            raise CallbackError(
                f"in callback '{self.name}' exception "
                f"'{err.__class__.__name__}' was raised, {err}")

    @classmethod
    def from_yaml(cls, name: str, data: str) -> Self:
        """
        :class:`MailCallback` can be also constructed from a YAML snippet.

        .. code:: yaml

            host: [SMTP_HOST]
            port: [SMTP_PORT]
            login: [SMTP_LOGIN]
            password: [SMTP_PASSWORD]
            security: [starttls | ssl]
            from: account@host.com
            to: account@anotherhost.com
            subject: Some subject with an attribute {subject}
            message: Some message with attributes {message} {date}
            attach:
                - patch/to/file1
                - patch/to/file2
            delay: 42

        :param name: unique identifier
        :param data: YAML snippet

        :return: new instance
        :rtype: MailCallback
        """
        super().from_yaml(name, data)

        try:
            parsed = yaml.safe_load(data)
        except yaml.YAMLError as err:
            raise CallbackSyntaxError(err)

        required_keys = [
            'host',
            'port',
            'login',
            'password',
            'security',
            'to',
            'subject',
            'from',
            'message'
        ]

        for required_key in required_keys:
            if required_key not in parsed:
                raise CallbackSyntaxError(
                    f"in callback '{name}' missing '{required_key}'")

        host = parsed['host']
        port = parsed['port']
        login = parsed['login']
        password = parsed['password']
        security = parsed['security']
        toaddr = parsed['to']
        subject = parsed['subject']
        fromaddr = parsed['from']
        message = parsed['message']

        attach: list[str] = []
        if 'attach' in parsed:
            if not isinstance(parsed['attach'], list):
                raise CallbackSyntaxError(
                    f"in callback '{name}' 'attach' not a list")

            for attachment in parsed['attach']:
                if not isinstance(attachment, str):
                    raise CallbackSyntaxError(
                        f"in callback '{name}' invalid attachment")

                attach.append(attachment)

        if not isinstance(host, str):
            raise CallbackSyntaxError(
                f"in callback '{name}' 'host' not a str")

        if not isinstance(login, str):
            raise CallbackSyntaxError(
                f"in callback '{name}' 'login' not a str")

        if not isinstance(password, str):
            raise CallbackSyntaxError(
                f"in callback '{name}' 'password' not a str")

        if not isinstance(security, str):
            raise CallbackSyntaxError(
                f"in callback '{name}' 'security' not a str")

        if not isinstance(toaddr, str):
            raise CallbackSyntaxError(
                f"in callback '{name}' 'to' not a str")

        if not isinstance(subject, str):
            raise CallbackSyntaxError(
                f"in callback '{name}' 'subject' not a str")

        if not isinstance(fromaddr, str):
            raise CallbackSyntaxError(
                f"in callback '{name}' 'from' not a str")

        if not isinstance(message, str):
            raise CallbackSyntaxError(
                f"in callback '{name}' 'message' not a str")

        if not isinstance(port, int):
            raise CallbackSyntaxError(
                f"in callback '{name}' 'port' not an int")

        if 'delay' in parsed:
            delay = parsed['delay']
            if not isinstance(delay, int):
                raise CallbackSyntaxError(
                    f"in callback '{name}' 'delay' not an int")
        else:
            delay = 0

        if security not in ['starttls', 'ssl']:
            raise CallbackSyntaxError(
                f"in callback '{name}' invalid "
                f"security '{security}' value")

        return cls(name, host, port, login, password, security,
                   toaddr, subject, fromaddr, message, attach, delay)



from loguru import logger
from abc import ABC, abstractmethod
import sys
import yaml
import systemd.journal


class LoggerSyntaxError(Exception):
    """
    Raised on logger syntax error
    """

    def __init__(self, message="logger syntax error"):
        self.message = message
        super().__init__(self.message)


class AbstractLogger(ABC):
    """
    Abstract class from which all logger classes are derived.

    The preferred way to instatiate a logger is from class
    method :func:`~from_yaml`.
    """

    @abstractmethod
    def __init__(self):
        self.check_imp_level()
        self.id = self.add_logger()
        logger.info(f'logger ({self.__class__}) initialized')

    @classmethod
    @abstractmethod
    def from_yaml(cls, data: str):
        """
        A class method for constructing a logger from a YAML document.

        :param data: yaml data

        :return: new instance
        """

        logger.debug(f'logger defined form yaml \n{data}')

    def add_logger(self):
        """
        Add logger with logging level.

        :param level: logging level (lower case)
        """

        return logger.add(self.handler,
                          format=self.format,
                          level=self.level.upper())

    def check_imp_level(self):
        imp_levels = [
            'trace',
            'debug',
            'info',
            'success',
            'warning',
            'error',
            'critical'
        ]

        if self.level not in imp_levels:
            raise LoggerSyntaxError(f"""\
            in {self.__class__} invalid logger level {self.level}
            """)


class StdErrLogger(AbstractLogger):
    """
    `stderr` logger.

    The preferred way to instatiate is from class
    method :func:`~from_yaml`.
    """

    def __init__(self, level: str = 'debug', format: str = (
        "<level>{level: <8}</level> | "
        "<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> | " # noqa
        "<level>{message}</level>"
    )):
        self.handler = sys.stderr
        self.level = level
        self.format = format
        super().__init__()

    @classmethod
    def from_yaml(cls, data: str):
        """
        A class method for constructing a stderr logger from a YAML document.

        :param data: yaml data

        :return: new instance
        """

        try:
            parsed = yaml.safe_load(data)
            super().from_yaml(data)
        except yaml.YAMLError as err:
            raise LoggerSyntaxError(err)

        if parsed is not None and 'level' in parsed:
            if isinstance(parsed['level'], str):
                return cls(parsed['level'])
            else:
                raise LoggerSyntaxError('in log_stderr expected '
                                        'level to be of type str')
        else:
            return cls()


class FileLogger(AbstractLogger):
    """
    Logger logging into a file

    The preferred way to instatiate is from class
    method :func:`~from_yaml`.
    """

    def __init__(self, path: str, level: str = 'debug',
                 format: str = ("<level>{level: <8}</level> | "
                                "<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> | " # noqa
                                "<level>{message}</level>")):
        self.handler = path
        self.path = path
        self.level = level

        self.format = format
        super().__init__()

    @classmethod
    def from_yaml(cls, data: str):
        """
        A class method for constructing a stderr logger from a YAML document.

        :param data: yaml data

        :return: new instance
        """

        try:
            parsed = yaml.safe_load(data)
            super().from_yaml(data)
        except yaml.YAMLError as err:
            raise LoggerSyntaxError(err)

        if parsed is not None and 'path' in parsed:
            path = parsed['path']
            if not isinstance(path, str):
                raise LoggerSyntaxError('in log_file expected '
                                        'path to be of type str')
        else:
            raise LoggerSyntaxError('in logger log_file path is missing')

        if parsed is not None and 'level' in parsed:
            level = parsed['level']
            if not isinstance(level, str):
                raise LoggerSyntaxError('in log_file expected '
                                        'level to be of type str')

            return cls(path, level)

        return cls(path)


class JournalLogger(AbstractLogger):
    """
    System journal logger.

    The preferred way to instatiate is from class
    method :func:`~from_yaml`.
    """

    def __init__(self, level: str = 'debug', format: str = (
        "<level>{level: <8}</level> | "
        "<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> | " # noqa
        "<level>{message}</level>"
    )):
        self.handler = systemd.journal.JournalHandler()
        self.level = level
        self.format = format
        super().__init__()

    @classmethod
    def from_yaml(cls, data: str):
        """
        A class method for constructing a system journal
        logger from a YAML document.

        :param data: yaml data

        :return: new instance
        """

        try:
            parsed = yaml.safe_load(data)
            super().from_yaml(data)
        except yaml.YAMLError as err:
            raise LoggerSyntaxError(f'in log_journal {err}')

        if parsed is not None and 'level' in parsed:
            level = parsed['level']
            if not isinstance(level, str):
                raise LoggerSyntaxError('in log_journal expected '
                                        'level to be of type str')
            else:
                return cls(level)

        return cls()



from yasmon.tasks import TaskRunner
from yasmon.tasks import TaskError
from yasmon.processor import YAMLProcessor
from yasmon.processor import YAMLSyntaxError
from yasmon.callbacks import CallbackSyntaxError
from yasmon.loggers import LoggerSyntaxError

from loguru import logger
import argparse
from pathlib import Path
from enum import IntEnum
import sys

logger.remove(0)


class ExitCodes(IntEnum):
    Success                         = 0  # noqa
    UnexpectedException             = 1  # noqa
    YAMLSyntaxError                 = 2  # noqa
    FileNotFoundError               = 3  # noqa
    TaskNotImplementedError         = 4  # noqa
    CallbackNotImplementedError     = 5  # noqa
    CallbackSyntaxError             = 6  # noqa
    TaskError                       = 7  # noqa
    LoggerSyntaxError               = 8  # noqa


def parse_args(*args):
    parser = argparse.ArgumentParser()
    parser.add_argument("--config", type=str,
                        help="yaml config file path",
                        default=str(Path(Path.home(),
                                         '.config',
                                         'yasmon',
                                         'config.yaml')))
    return parser.parse_args(args)


def setup_default_logger():
    journal_logger_format = (
        "<level>{level: <8}</level> | "
        "<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> | "
        "<level>{message}</level>"
    )
    return logger.add(sys.stderr, format=journal_logger_format, level='DEBUG')


def execute(args, default_logger_id):
    try:
        processor = YAMLProcessor()
        processor.load_file(args.config)

        loggers = processor.add_loggers()
        num_usr_loggers = len(loggers)
        if num_usr_loggers > 0:
            logger.remove(default_logger_id)
            logger.info(f'there are {num_usr_loggers} user loggers defined: '
                        'default stderr logger removed')
        else:
            logger.warning('there are no user loggers defined: '
                           'keeping default stderr logger')

        callbacks = processor.get_callbacks()
        tasks = processor.get_tasks(callbacks)
        runner = TaskRunner(tasks)
        runner.loop.run_until_complete(runner())
        return ExitCodes.Success
    except YAMLSyntaxError as err:
        logger.error(f'{err.__class__.__name__}: {err}')
        return ExitCodes.YAMLSyntaxError
    except FileNotFoundError as err:
        logger.error(f'{err.__class__.__name__}: {err}')
        return ExitCodes.FileNotFoundError
    except CallbackSyntaxError as err:
        logger.error(f'{err.__class__.__name__}: {err}')
        return ExitCodes.CallbackSyntaxError
    except TaskError as err:
        logger.error(f'{err.__class__.__name__}: {err}')
        return ExitCodes.TaskError
    except LoggerSyntaxError as err:
        logger.error(f'{err.__class__.__name__}: {err}')
        return ExitCodes.LoggerSyntaxError
    except Exception as err:
        logger.error(f'Unexpected {err.__class__.__name__}: {err}')
        return ExitCodes.UnexpectedException


def main():
    default_logger_id = setup_default_logger()
    args = parse_args(*sys.argv[1:])
    return execute(args, default_logger_id)


if __name__ == '__main__':
    sys.exit(main())



from yasmon.processor import YAMLProcessor, YAMLSyntaxError
from yasmon.callbacks import CallbackDict, CallbackSyntaxError
from yasmon.callbacks import CallbackNotImplementedError
from yasmon.tasks import TaskNotImplementedError
from yasmon.loggers import LoggerSyntaxError

from loguru import logger
import unittest
from unittest import mock


class YAMLProcessorTest(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(YAMLProcessorTest, self).__init__(*args, **kwargs)
        self.proc = YAMLProcessor()

    def test_load_file_raise_exceptions(self):
        """
        Test if YAMLProcessor.load_file() raises appropriate exceptions.
        """

        fun = self.proc.load_file

        # FileNotFoundError
        self.assertRaises(FileNotFoundError, fun, "tests/assets/notafile")
        self.assertRaises(FileNotFoundError, fun, "tests/asset/config.yaml")

        # OSError
        with mock.patch('builtins.open') as mock_oserror:
            mock_oserror.side_effect = OSError
            self.assertRaises(OSError, fun, 'tests/assets/config.yaml')

        # Exception
        with mock.patch('builtins.open') as mock_oserror:
            mock_oserror.side_effect = Exception
            self.assertRaises(Exception, fun, 'tests/assets/config.yaml')

        # PermissionError
        with mock.patch('builtins.open') as mock_oserror:
            mock_oserror.side_effect = PermissionError
            self.assertRaises(PermissionError, fun, 'tests/assets/config.yaml')

        # YAMLSyntaxError
        self.assertRaises(YAMLSyntaxError, fun, 'tests/assets/invalid.yaml')

        # empty config file
        path = 'tests/assets/empty.yaml'
        self.assertRaises(YAMLSyntaxError, self.proc.load_file, path)

    def test_load_document_raise_exceptions(self):
        """
        Test if YAMLProcessor.load_document() raises appropriate exceptions.
        """

        # general syntax error
        test_yaml = """
        key: ][
        """
        self.assertRaises(YAMLSyntaxError, self.proc.load_document, test_yaml)

        # config is empty
        test_yaml = """
        """
        self.assertRaises(YAMLSyntaxError, self.proc.load_document, test_yaml)

    def test_add_logger_implemented_loggers(self):
        """
        Test if YAMLProcessor.add_loggers() can process implemented loggers.
        """

        # log_stderr
        test_yaml = """
        log_stderr:
        """
        self.proc.load_document(test_yaml)
        assert len(self.proc.add_loggers()) == 1
        logger.remove()

        # log_stderr with level info
        test_yaml = """
        log_stderr:
            level: info
        """
        self.proc.load_document(test_yaml)
        assert len(self.proc.add_loggers()) == 1
        assert self.proc.data['log_stderr']['level'] == 'info'
        logger.remove()

        # log_journal
        test_yaml = """
        log_journal:
        """
        self.proc.load_document(test_yaml)
        assert len(self.proc.add_loggers()) == 1
        logger.remove()

        # log_journal with level debug
        test_yaml = """
        log_journal:
            level: debug
        """
        self.proc.load_document(test_yaml)
        assert len(self.proc.add_loggers()) == 1
        assert self.proc.data['log_journal']['level'] == 'debug'
        logger.remove()

        # log_file
        test_yaml = """
        log_file:
            path: tests/assets/tmp/logfile
        """
        self.proc.load_document(test_yaml)
        assert self.proc.data['log_file']['path'] == 'tests/assets/tmp/logfile'
        assert len(self.proc.add_loggers()) == 1
        logger.remove()

        # log_file with level debug
        test_yaml = """
        log_file:
            path: tests/assets/tmp/logfile
            level: error
        """
        self.proc.load_document(test_yaml)
        assert len(self.proc.add_loggers()) == 1
        assert self.proc.data['log_file']['level'] == 'error'
        assert self.proc.data['log_file']['path'] == 'tests/assets/tmp/logfile'
        logger.remove()

    def test_add_logger_raise_exceptions(self):
        """
        Test if YAMLProcessor.add_loggers() raises appropriate exceptions.
        """

        # invalid level
        test_yaml = """
        log_stderr:
            level: []
        """
        self.proc.load_document(test_yaml)
        self.assertRaises(LoggerSyntaxError, self.proc.add_loggers)
        logger.remove()

        # invalid level
        test_yaml = """
        log_journal:
            level: []
        """
        self.proc.load_document(test_yaml)
        self.assertRaises(LoggerSyntaxError, self.proc.add_loggers)
        logger.remove()

        # invalid level
        test_yaml = """
        log_file:
            path: /tmp/log
            level: []
        """
        self.proc.load_document(test_yaml)
        self.assertRaises(LoggerSyntaxError, self.proc.add_loggers)
        logger.remove()

        # invalid path
        test_yaml = """
        log_file:
            path: []
        """
        self.proc.load_document(test_yaml)
        self.assertRaises(LoggerSyntaxError, self.proc.add_loggers)
        logger.remove()

        # invalid level (not implemented)
        test_yaml = """
        log_stderr:
            level: notalevel
        """
        self.proc.load_document(test_yaml)
        self.assertRaises(LoggerSyntaxError, self.proc.add_loggers)

        # invalid level (not implemented)
        test_yaml = """
        log_journal:
            level: notalevel
        """
        self.proc.load_document(test_yaml)
        self.assertRaises(LoggerSyntaxError, self.proc.add_loggers)

        # OSError
        test_yaml = """
        log_file:
            path: /tmp/log
        """
        self.proc.load_document(test_yaml)
        with mock.patch('builtins.open') as mock_oserror:
            mock_oserror.side_effect = OSError
            self.assertRaises(OSError, self.proc.add_loggers)

        # missing path for log_file
        test_yaml = """
        log_file:
        """
        self.proc.load_document(test_yaml)
        self.assertRaises(LoggerSyntaxError, self.proc.add_loggers)

        # adds all defined loggers
        test_yaml = """
        log_stderr:
        log_journal:
        log_file:
            path: /tmp/test_add_logger_file.log
        """
        self.proc.load_document(test_yaml)
        assert len(self.proc.add_loggers()) == 3
        assert self.proc.data['log_file']['path'] == '/tmp/test_add_logger_file.log' # noqa
        logger.remove()

    def test_get_tasks_raise_exceptions(self):
        """
        Test if YAMLProcessor.get_tasks() raises appropriate exceptions.
        """

        fun = self.proc.get_tasks

        # tasks not defined
        test_yaml = """
        key:
        """
        self.proc.load_document(test_yaml)
        self.assertRaises(YAMLSyntaxError, fun, CallbackDict())

        # tasks are not a dictionary
        test_yaml = """
        tasks:
            - sometask
        """
        self.proc.load_document(test_yaml)
        self.assertRaises(YAMLSyntaxError, fun, CallbackDict())

        # task is missing callbacks
        test_yaml = """
        tasks:
            sometask:
                type: watchfiles
                paths:
                    - /tmp/
                changes:
                    - added
        """
        self.proc.load_document(test_yaml)
        self.assertRaises(YAMLSyntaxError, fun, CallbackDict())

        # task data (sometask) is not a dictionary
        test_yaml = """
        tasks:
            sometask:
                - type: watchfiles
        """
        self.proc.load_document(test_yaml)
        self.assertRaises(YAMLSyntaxError, fun, CallbackDict())

        # task is not implemented
        test_yaml = """
        callbacks:
            callback0:
                type: shell
                command: exit 0
        tasks:
            sometask:
                type: notimplemented
                changes:
                    - added
                    - modified
                    - deleted
                paths:
                    - /tmp/
                callbacks:
                    - callback0
        """
        self.proc.load_document(test_yaml)
        callbacks = self.proc.get_callbacks()
        self.assertRaises(TaskNotImplementedError, fun, callbacks)

    def test_get_callbacks_raise_exceptions(self):
        """
        Test if YAMLProcessor.get_callbacks() raises appropriate exceptions.
        """

        fun = self.proc.get_callbacks

        # callbacks not defined
        test_yaml = """
        tasks:
        """
        self.proc.load_document(test_yaml)
        self.assertRaises(YAMLSyntaxError, fun)

        # callbacks not a dictionary
        test_yaml = """
        callbacks:
            - callback0
        """
        self.proc.load_document(test_yaml)
        self.assertRaises(YAMLSyntaxError, fun)

        # callback data (callback0) is not a dictionary
        test_yaml = """
        callbacks:
            callback0:
                - type: watchfiles
        """
        self.proc.load_document(test_yaml)
        self.assertRaises(YAMLSyntaxError, fun)

        # callback is not implemented
        test_yaml = """
        callbacks:
            callback0:
                type: notimplemented
        """
        self.proc.load_document(test_yaml)
        self.assertRaises(CallbackNotImplementedError, fun)

        # invalid callback syntax
        test_yaml = """
        callbacks:
            callback0:
                type: logger
                level: notdefined
        """
        self.proc.load_document(test_yaml)
        self.assertRaises(CallbackSyntaxError, fun)


if __name__ == '__main__':
    unittest.main(verbosity=2)



from . import processor
from . import callbacks
from . import tasks
from . import loggers
from . import cli

__all__ = [
    'processor',
    'callbacks',
    'tasks',
    'loggers',
    'cli',
]

__name__ = 'yasmon'



from yasmon import cli
from loguru import logger

import unittest
from unittest import mock


class CLITest(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(CLITest, self).__init__(*args, **kwargs)

    def test_parse_args_value_assignment(self):
        """
        Test cli.parse_args() for assignment of options.
        """

        config_file = 'tests/assets/cli/config.yaml'
        args = cli.parse_args('--config', config_file)
        assert args.config == config_file

    def test_execute_ExitCode_Success(self):
        """
        Test if cli.execute() returns ExitCodes.Success

        This exit code is returned in the case that no errors occured
        during operation.
        """

        default_logger_id = cli.setup_default_logger()
        args = cli.parse_args('--config', 'tests/assets/config.yaml')
        code = cli.execute(args, default_logger_id)
        assert code == cli.ExitCodes.Success
        logger.remove()

    def test_execute_ExitCode_FileNotFoundError(self):
        """
        Test if cli.execute() returns ExitCodes.FileNotFoundError

        This exit code should be returned if the path to the config file
        does not exist.
        """

        default_logger_id = cli.setup_default_logger()
        args = cli.parse_args('--config', 'invalid/path/to/config')
        code = cli.execute(args, default_logger_id)
        assert code == cli.ExitCodes.FileNotFoundError
        logger.remove()

    def test_execute_ExitCode_YAMLSyntaxError(self):
        """
        Test if cli.execute() returns ExitCodes.YAMLSyntaxError

        This exit code should be returned if the config syntax
        is invalid. This includes typical YAML Syntax errors as well as
        Yasmon specific syntax errors like missing callbacks and tasks.
        """

        default_logger_id = cli.setup_default_logger()
        args = cli.parse_args('--config', 'tests/assets/invalid.yaml')
        code = cli.execute(args, default_logger_id)
        assert code == cli.ExitCodes.YAMLSyntaxError
        logger.remove()

    def test_execute_catches_exceptions(self):
        """
        Test if cli.execute() catches all necessary exceptions.

        This only tests if all appropriate ExitCodes are returned.
        Actual rasing of exceptions is tested elsewhere.
        """

        # cli.ExitCodes.CallbackSyntaxError
        default_logger_id = cli.setup_default_logger()
        args = cli.parse_args(
            '--config',
            'tests/assets/CallbackSyntaxError.yaml')
        code = cli.execute(args, default_logger_id)
        assert code == cli.ExitCodes.CallbackSyntaxError
        logger.remove()

        # cli.ExitCodes.TaskError
        default_logger_id = cli.setup_default_logger()
        args = cli.parse_args(
            '--config',
            'tests/assets/TaskError.yaml')
        code = cli.execute(args, default_logger_id)
        assert code == cli.ExitCodes.TaskError
        logger.remove()

        # cli.ExitCodes.LoggerSyntaxError
        default_logger_id = cli.setup_default_logger()
        args = cli.parse_args(
            '--config',
            'tests/assets/LoggerSyntaxError.yaml')
        code = cli.execute(args, default_logger_id)
        assert code == cli.ExitCodes.LoggerSyntaxError
        logger.remove()

        # cli.ExitCodes.UnexpectedException
        with mock.patch('yasmon.processor.YAMLProcessor.__init__') as mck:
            mck.side_effect = Exception
            default_logger_id = cli.setup_default_logger()
            args = cli.parse_args(
                '--config',
                'tests/assets/config.yaml')
            code = cli.execute(args, default_logger_id)
            assert code == cli.ExitCodes.UnexpectedException
            logger.remove()


if __name__ == '__main__':
    unittest.main(verbosity=2)



from yasmon.processor import YAMLProcessor
from yasmon.tasks import TaskRunner
from yasmon.callbacks import ShellCallback
from yasmon.callbacks import LoggerCallback
from yasmon.callbacks import MailCallback
from yasmon.callbacks import CallbackError
from yasmon.callbacks import CallbackSyntaxError
from yasmon.callbacks import CallbackCircularAttributeError
from yasmon.callbacks import CallbackAttributeError

import unittest
from unittest import mock
import subprocess
import asyncio
import time
import os
import random
import string
import poplib
import email
from email.message import EmailMessage
import filecmp
import mimetypes


class ShellCallbackTest(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(ShellCallbackTest, self).__init__(*args, **kwargs)
        self.proc = YAMLProcessor()
        self.input_producer_script = 'tests/assets/watchfiles_test.sh'
        self.input_producer = None

    def start_input_producer(self):
        self.input_producer = subprocess.Popen(self.input_producer_script,
                                               stdout=subprocess.PIPE,
                                               stderr=subprocess.PIPE)

        # wait a moment for input producer (just to be safe)
        time.sleep(1)

    def stop_input_producer(self):
        if self.input_producer:
            self.input_producer.kill()
            self.input_producer.wait()
            self.input_producer.communicate()
            self.input_producer = None

    def test_from_yaml_assignment(self):
        """
        Test ShellCallback.from_yaml() for value assignment.
        """

        test_yaml = """
            type: shell
            command: somecommand
        """
        callback = ShellCallback.from_yaml('name', test_yaml)
        assert callback.name == 'name'
        assert callback.cmd == 'somecommand'

    def test_ShellCallback_raise_exceptions(self):
        """
        Test ShellCallback.from_yaml() for proper exceptions.
        """

        # general yaml error
        test_yaml = """
            type: shell
            command: ]]
        """
        fun = ShellCallback.from_yaml
        self.assertRaises(CallbackSyntaxError, fun, 'name', test_yaml)

        # command not defined
        test_yaml = """
            type: shell
        """
        fun = ShellCallback.from_yaml
        self.assertRaises(CallbackSyntaxError, fun, 'name', test_yaml)

        # command not a string
        test_yaml = """
            type: shell
            command: []
        """
        fun = ShellCallback.from_yaml
        self.assertRaises(CallbackSyntaxError, fun, 'name', test_yaml)

    def test_call_success(self):
        """
        Test ShellCallback.__call__() with stdout for success.
        """

        # stdout (command success)
        test_yaml = """
        callbacks:
            callback0:
                type: shell
                command: echo {myattr}; return 0
        tasks:
            watchfilestask:
                type: watchfiles
                changes:
                    - modified
                paths:
                    - tests/assets/tmp/watchfiles_call_test
                callbacks:
                    - callback0
                attrs:
                    myattr: somevalue
        """
        self.start_input_producer()
        self.proc.load_document(test_yaml)
        callbacks = self.proc.get_callbacks()
        tasks = self.proc.get_tasks(callbacks)
        runner = TaskRunner(tasks, testenv=True)
        try:
            runner.loop.run_until_complete(runner())
        except Exception:
            self.fail()
        self.stop_input_producer()

        # stderr (command failure)
        test_yaml = """
        callbacks:
            callback0:
                type: shell
                command: ">&2 echo {myattr}; return 1"
        tasks:
            watchfilestask:
                type: watchfiles
                changes:
                    - modified
                paths:
                    - tests/assets/tmp/watchfiles_call_test
                callbacks:
                    - callback0
                attrs:
                    myattr: somevalue
        """
        self.start_input_producer()
        self.proc.load_document(test_yaml)
        callbacks = self.proc.get_callbacks()
        tasks = self.proc.get_tasks(callbacks)
        runner = TaskRunner(tasks, testenv=True)
        try:
            runner.loop.run_until_complete(runner())
        except Exception:
            self.fail()
        self.stop_input_producer()


class LoggerCallbackTest(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(LoggerCallbackTest, self).__init__(*args, **kwargs)
        self.proc = YAMLProcessor()
        self.input_producer_script = 'tests/assets/watchfiles_test.sh'
        self.input_producer = None

    def start_input_producer(self):
        self.input_producer = subprocess.Popen(self.input_producer_script,
                                               stdout=subprocess.PIPE,
                                               stderr=subprocess.PIPE)

        # wait a moment for input producer (just to be safe)
        time.sleep(1)

    def stop_input_producer(self):
        if self.input_producer:
            self.input_producer.kill()
            self.input_producer.wait()
            self.input_producer.communicate()
            self.input_producer = None

    def test_from_yaml_assignment(self):
        """
        Test LoggerCallback.from_yaml() for value assignment.
        """

        test_yaml = """
            level: trace
            message: some message
        """
        callback = LoggerCallback.from_yaml('name', test_yaml)
        assert callback.name == 'name'
        assert callback.level == 'trace'
        assert callback.message == 'some message'

    def test_from_yaml_raise_exceptions(self):
        """
        Test LoggerCallback.from_yaml() for proper exceptions.
        """

        fun = LoggerCallback.from_yaml

        # general yaml error
        test_yaml = """
            level: trace
            message: ]]
        """
        self.assertRaises(CallbackSyntaxError, fun, 'name', test_yaml)

        # missing level
        test_yaml = """
            message: some message
        """
        self.assertRaises(CallbackSyntaxError, fun, 'name', test_yaml)

        # invalid level type
        test_yaml = """
            level: []
            message: some message
        """
        self.assertRaises(CallbackSyntaxError, fun, 'name', test_yaml)

        # invalid level
        test_yaml = """
            level: INVALID
            message: some message
        """
        self.assertRaises(CallbackSyntaxError, fun, 'name', test_yaml)

        # missing message
        test_yaml = """
            level: info
        """
        self.assertRaises(CallbackSyntaxError, fun, 'name', test_yaml)

        # invalid message type
        test_yaml = """
            level: info
            message: []
        """
        self.assertRaises(CallbackSyntaxError, fun, 'name', test_yaml)

    def test_call_success(self):
        """
        Test LoggerCallback.__call__() for success.
        """

        test_yaml = """
        callbacks:
            logger0:
                type: logger
                level: info
                message: "{myattr}"
        tasks:
            watchfilestask:
                type: watchfiles
                changes:
                    - modified
                paths:
                    - tests/assets/tmp/watchfiles_call_test
                callbacks:
                    - logger0
                attrs:
                    myattr: somevalue
        """
        self.start_input_producer()
        self.proc.load_document(test_yaml)
        callbacks = self.proc.get_callbacks()
        tasks = self.proc.get_tasks(callbacks)
        runner = TaskRunner(tasks, testenv=True)
        try:
            runner.loop.run_until_complete(runner())
        except Exception:
            self.fail()
        self.stop_input_producer()

    def test_call_raise_exceptions(self):
        """
        Test LoggerCallback.__call__() for proper exceptions
        """

        # CallbackCircularAttributeError
        test_yaml = """
        callbacks:
            logger0:
                type: logger
                level: info
                message: "{myattr}"
        tasks:
            watchfilestask:
                type: watchfiles
                changes:
                    - modified
                paths:
                    - tests/assets/tmp/watchfiles_call_test
                callbacks:
                    - logger0
                attrs:
                    myattr: this {INVALID} attribute
                    INVALID: is {myattr} circular
        """
        self.start_input_producer()
        self.proc.load_document(test_yaml)
        callbacks = self.proc.get_callbacks()
        tasks = self.proc.get_tasks(callbacks)
        runner = TaskRunner(tasks, testenv=True)
        fun = runner.loop.run_until_complete
        self.assertRaises(CallbackCircularAttributeError, fun, runner())
        self.stop_input_producer()

        # CallbackAttributeError
        test_yaml = """
        callbacks:
            logger0:
                type: logger
                level: info
                message: "{myattr}"
        tasks:
            watchfilestask:
                type: watchfiles
                changes:
                    - modified
                paths:
                    - tests/assets/tmp/watchfiles_call_test
                callbacks:
                    - logger0
                attrs:
                    INVALID: myattr is not defined
        """
        self.start_input_producer()
        self.proc.load_document(test_yaml)
        callbacks = self.proc.get_callbacks()
        tasks = self.proc.get_tasks(callbacks)
        runner = TaskRunner(tasks, testenv=True)
        fun = runner.loop.run_until_complete
        self.assertRaises(CallbackAttributeError, fun, runner())
        self.stop_input_producer()


class MailCallbackTest(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(MailCallbackTest, self).__init__(*args, **kwargs)
        self.SMTP_HOST = os.environ.get('YASMON_SMTP_HOST')
        self.POP3_HOST = os.environ.get('YASMON_POP3_HOST')
        self.SMTP_LOGIN = os.environ.get('YASMON_SMTP_LOGIN')
        self.SMTP_PASSWORD = os.environ.get('YASMON_SMTP_SECRET')

    def test_from_yaml_assignment(self):
        """
        Test MailCallback.from_yaml() for value assignment.
        """

        test_yaml = f"""
            host: "{self.SMTP_HOST}"
            port: 587
            login: "{self.SMTP_LOGIN}"
            password: "{self.SMTP_PASSWORD}"
            security: starttls
            from: "{{from}}"
            to: "{{to}}"
            subject: Notification {{subject}}
            message: "{{message}}"
            attach:
                - path/to/file1
                - path/to/file2
            delay: 10
        """
        mail = MailCallback.from_yaml('name', test_yaml)
        assert mail.host == self.SMTP_HOST
        assert mail.port == 587
        assert mail.login == self.SMTP_LOGIN
        assert mail.password == self.SMTP_PASSWORD
        assert mail.security == 'starttls'
        assert mail.fromaddr == '{from}'
        assert mail.toaddr == '{to}'
        assert mail.subject == 'Notification {subject}'
        assert mail.message == '{message}'
        assert mail.attach == ['path/to/file1', 'path/to/file2']
        assert mail.delay == 10

    def test_from_yaml_raise_exceptions(self):
        """
        Test MailCallback.from_yaml() for proper exceptions.
        """

        fun = MailCallback.from_yaml

        # general yaml error
        test_yaml = """
            message: ][
        """
        self.assertRaises(CallbackSyntaxError, fun, 'name', test_yaml)

        # missing host
        test_yaml = f"""
            port: 587
            login: "{self.SMTP_LOGIN}"
            password: "{self.SMTP_PASSWORD}"
            security: starttls
            from: "{{from}}"
            to: "{{to}}"
            subject: "Notification: {{subject}}"
            message: "{{message}}"
        """
        with self.assertRaises(CallbackSyntaxError) as context:
            fun('name', test_yaml)
        err = context.exception
        assert str(err) == "in callback 'name' missing 'host'"

        # missing port
        test_yaml = f"""
            host: "{self.SMTP_HOST}"
            login: "{self.SMTP_LOGIN}"
            password: "{self.SMTP_PASSWORD}"
            security: starttls
            from: "{{from}}"
            to: "{{to}}"
            subject: "Notification: {{subject}}"
            message: "{{message}}"
        """
        with self.assertRaises(CallbackSyntaxError) as context:
            fun('name', test_yaml)
        err = context.exception
        assert str(err) == "in callback 'name' missing 'port'"

        # missing login
        test_yaml = f"""
            host: "{self.SMTP_HOST}"
            port: 587
            password: "{self.SMTP_PASSWORD}"
            security: starttls
            from: "{{from}}"
            to: "{{to}}"
            subject: "Notification: {{subject}}"
            message: "{{message}}"
        """
        with self.assertRaises(CallbackSyntaxError) as context:
            fun('name', test_yaml)
        err = context.exception
        assert str(err) == "in callback 'name' missing 'login'"

        # missing password
        test_yaml = f"""
            host: "{self.SMTP_HOST}"
            port: 587
            login: "{self.SMTP_LOGIN}"
            security: starttls
            from: "{{from}}"
            to: "{{to}}"
            subject: "Notification: {{subject}}"
            message: "{{message}}"
        """
        with self.assertRaises(CallbackSyntaxError) as context:
            fun('name', test_yaml)
        err = context.exception
        assert str(err) == "in callback 'name' missing 'password'"

        # missing security
        test_yaml = f"""
            host: "{self.SMTP_HOST}"
            port: 587
            login: "{self.SMTP_LOGIN}"
            password: "{self.SMTP_PASSWORD}"
            from: "{{from}}"
            to: "{{to}}"
            subject: "Notification: {{subject}}"
            message: "{{message}}"
        """
        with self.assertRaises(CallbackSyntaxError) as context:
            fun('name', test_yaml)
        err = context.exception
        assert str(err) == "in callback 'name' missing 'security'"

        # missing subject
        test_yaml = f"""
            host: "{self.SMTP_HOST}"
            port: 587
            login: "{self.SMTP_LOGIN}"
            password: "{self.SMTP_PASSWORD}"
            security: starttls
            from: "{{from}}"
            to: "{{to}}"
            message: "{{message}}"
        """
        with self.assertRaises(CallbackSyntaxError) as context:
            fun('name', test_yaml)
        err = context.exception
        assert str(err) == "in callback 'name' missing 'subject'"

        # missing to address
        test_yaml = f"""
            host: "{self.SMTP_HOST}"
            port: 587
            login: "{self.SMTP_LOGIN}"
            password: "{self.SMTP_PASSWORD}"
            subject: "Notification: {{subject}}"
            from: {{from}}
            security: starttls
            message: "{{message}}"
        """
        with self.assertRaises(CallbackSyntaxError) as context:
            fun('name', test_yaml)
        err = context.exception
        assert str(err) == "in callback 'name' missing 'to'"

        # missing from
        test_yaml = f"""
            host: "{self.SMTP_HOST}"
            port: 587
            login: "{self.SMTP_LOGIN}"
            password: "{self.SMTP_PASSWORD}"
            security: starttls
            to: "{{to}}"
            subject: "Notification: {{subject}}"
            message: "{{message}}"
        """
        with self.assertRaises(CallbackSyntaxError) as context:
            fun('name', test_yaml)
        err = context.exception
        assert str(err) == "in callback 'name' missing 'from'"

        # missing message
        test_yaml = f"""
            host: "{self.SMTP_HOST}"
            port: 587
            login: "{self.SMTP_LOGIN}"
            password: "{self.SMTP_PASSWORD}"
            subject: "Notification: {{subject}}"
            from: {{from}}
            to: "{{to}}"
            security: starttls
        """
        with self.assertRaises(CallbackSyntaxError) as context:
            fun('name', test_yaml)
        err = context.exception
        assert str(err) == "in callback 'name' missing 'message'"

        # port is not and int
        test_yaml = f"""
            host: "{self.SMTP_HOST}"
            port: INVALID
            login: "{self.SMTP_LOGIN}"
            password: "{self.SMTP_PASSWORD}"
            from: "{{from}}"
            to: "{{to}}"
            subject: "Notification: {{subject}}"
            security: starttls
            message: "{{message}}"
        """
        with self.assertRaises(CallbackSyntaxError) as context:
            fun('name', test_yaml)
        err = context.exception
        assert str(err) == "in callback 'name' 'port' not an int"

        # delay is not and int
        test_yaml = f"""
            host: "{self.SMTP_HOST}"
            port: 587
            delay: INVALID
            login: "{self.SMTP_LOGIN}"
            password: "{self.SMTP_PASSWORD}"
            from: "{{from}}"
            to: "{{to}}"
            subject: "Notification: {{subject}}"
            security: starttls
            message: "{{message}}"
        """
        with self.assertRaises(CallbackSyntaxError) as context:
            fun('name', test_yaml)
        err = context.exception
        assert str(err) == "in callback 'name' 'delay' not an int"

        # host not a string
        test_yaml = f"""
            host: []
            port: 587
            login: "{self.SMTP_LOGIN}"
            password: "{self.SMTP_PASSWORD}"
            security: starttls
            from: "{{from}}"
            to: "{{to}}"
            subject: "Notification: {{subject}}"
            message: "{{message}}"
        """
        with self.assertRaises(CallbackSyntaxError) as context:
            fun('name', test_yaml)
        err = context.exception
        assert str(err) == "in callback 'name' 'host' not a str"

        # login not a string
        test_yaml = f"""
            host: "{self.SMTP_HOST}"
            port: 587
            login: []
            password: "{self.SMTP_PASSWORD}"
            security: starttls
            from: "{{from}}"
            to: {{to}}
            subject: "Notification: {{subject}}"
            message: "{{message}}"
        """
        with self.assertRaises(CallbackSyntaxError) as context:
            fun('name', test_yaml)
        err = context.exception
        assert str(err) == "in callback 'name' 'login' not a str"

        # password not a string
        test_yaml = f"""
            host: "{self.SMTP_HOST}"
            port: 587
            login: "{self.SMTP_LOGIN}"
            password: []
            security: starttls
            from: "{{from}}"
            to: "{{to}}"
            subject: "Notification: {{subject}}"
            message: "{{message}}"
        """
        with self.assertRaises(CallbackSyntaxError) as context:
            fun('name', test_yaml)
        err = context.exception
        assert str(err) == "in callback 'name' 'password' not a str"

        # security not a string
        test_yaml = f"""
            host: "{self.SMTP_HOST}"
            port: 587
            login: "{self.SMTP_LOGIN}"
            password: "{self.SMTP_PASSWORD}"
            security: []
            from: "{{from}}"
            to: "{{to}}"
            subject: "Notification: {{subject}}"
            message: "{{message}}"
        """
        with self.assertRaises(CallbackSyntaxError) as context:
            fun('name', test_yaml)
        err = context.exception
        assert str(err) == "in callback 'name' 'security' not a str"

        # subject not a string
        test_yaml = f"""
            host: "{self.SMTP_HOST}"
            port: 587
            login: "{self.SMTP_LOGIN}"
            password: "{self.SMTP_PASSWORD}"
            security: starttls
            from: "{{from}}"
            to: "{{to}}"
            subject: []
            message: "{{message}}"
        """
        with self.assertRaises(CallbackSyntaxError) as context:
            fun('name', test_yaml)
        err = context.exception
        assert str(err) == "in callback 'name' 'subject' not a str"

        # from not a string
        test_yaml = f"""
            host: "{self.SMTP_HOST}"
            port: 587
            login: "{self.SMTP_LOGIN}"
            password: "{self.SMTP_PASSWORD}"
            security: starttls
            from: []
            to: "{{to}}"
            subject: "Notification: {{subject}}"
            message: "{{message}}"
        """
        with self.assertRaises(CallbackSyntaxError) as context:
            fun('name', test_yaml)
        err = context.exception
        assert str(err) == "in callback 'name' 'from' not a str"

        # to not a string
        test_yaml = f"""
            host: "{self.SMTP_HOST}"
            port: 587
            login: "{self.SMTP_LOGIN}"
            password: "{self.SMTP_PASSWORD}"
            security: starttls
            from: "{{from}}"
            to: []
            subject: "Notification: {{subject}}"
            message: "{{message}}"
        """
        with self.assertRaises(CallbackSyntaxError) as context:
            fun('name', test_yaml)
        err = context.exception
        assert str(err) == "in callback 'name' 'to' not a str"

        # message not a string
        test_yaml = f"""
            host: "{self.SMTP_HOST}"
            port: 587
            login: "{self.SMTP_LOGIN}"
            password: "{self.SMTP_PASSWORD}"
            security: starttls
            from: "{{from}}"
            to: "{{to}}"
            subject: "Notification: {{subject}}"
            message: []
        """
        with self.assertRaises(CallbackSyntaxError) as context:
            fun('name', test_yaml)
        err = context.exception
        assert str(err) == "in callback 'name' 'message' not a str"

        # invalid security
        test_yaml = f"""
            host: "{self.SMTP_HOST}"
            port: 587
            login: "{self.SMTP_LOGIN}"
            password: "{self.SMTP_PASSWORD}"
            security: WRONG
            from: "{{from}}"
            to: "{{to}}"
            subject: "Notification: {{subject}}"
            message: some message
        """
        with self.assertRaises(CallbackSyntaxError) as context:
            fun('name', test_yaml)
        err = context.exception
        assert str(err) == "in callback 'name' invalid security 'WRONG' value"

        # invalid attach
        test_yaml = f"""
            host: "{self.SMTP_HOST}"
            port: 587
            login: "{self.SMTP_LOGIN}"
            password: "{self.SMTP_PASSWORD}"
            security: WRONG
            from: "{{from}}"
            to: "{{to}}"
            subject: "Notification: {{subject}}"
            message: some message
            attach:
                file0:
                file1:
        """
        with self.assertRaises(CallbackSyntaxError) as context:
            fun('name', test_yaml)
        err = context.exception
        assert str(err) == "in callback 'name' 'attach' not a list"

        # attach list entry not a string
        test_yaml = f"""
            host: "{self.SMTP_HOST}"
            port: 587
            login: "{self.SMTP_LOGIN}"
            password: "{self.SMTP_PASSWORD}"
            security: WRONG
            from: "{{from}}"
            to: "{{to}}"
            subject: "Notification: {{subject}}"
            message: some message
            attach:
                - []
                - /path/to/file
        """
        with self.assertRaises(CallbackSyntaxError) as context:
            fun('name', test_yaml)
        err = context.exception
        assert str(err) == "in callback 'name' invalid attachment"

    def test_call(self):
        """
        Test MailCallback.__call__() to run through.
        """

        # starttls security
        subject_ = ''.join(random.choices(string.ascii_lowercase, k=42))
        message_ = ''.join(random.choices(string.ascii_lowercase, k=42))
        test_yaml = f"""
        callbacks:
            mail0:
                type: mail
                host: "{self.SMTP_HOST}"
                port: 587
                login: "{self.SMTP_LOGIN}"
                password: "{self.SMTP_PASSWORD}"
                security: starttls
                from: "{self.SMTP_LOGIN}"
                to: "{self.SMTP_LOGIN}"
                subject: "{subject_}"
                message: "{message_}{{myattr}}"
                attach:
                    - tests/assets/mailcallback_test_file.png
                delay: 1
        tasks:
            watchfiles:
                type: watchfiles
                changes:
                    - deleted
                paths:
                    - tests/assets/
                callbacks: []
                attrs:
                    myattr: somevalue
        """
        proc = YAMLProcessor()
        proc.load_document(test_yaml)
        callbacks = proc.get_callbacks()
        tasks = proc.get_tasks(callbacks)

        mail = callbacks['mail0']
        task = tasks[0]

        # send out random message
        try:
            loop = asyncio.new_event_loop()
            loop.run_until_complete(mail(task, task.attrs))
            loop.close()
        except Exception:
            self.fail()

        time.sleep(30)  # just to be safe

        # retrive all messages from server and look for our random message
        server = poplib.POP3_SSL(self.POP3_HOST)
        server.user(self.SMTP_LOGIN)
        server.pass_(self.SMTP_PASSWORD)
        content_ = message_ + 'somevalue\n'
        message_correct = False
        content_correct = False
        try:
            num_messages = len(server.list()[1])
            for k in range(1, num_messages + 1):
                envelope = b'\n'.join(server.retr(k)[1])
                message = email.message_from_bytes(
                        envelope, _class=EmailMessage)

                subject = message['Subject']
                subject_correct = (subject == subject_)

                if message.is_multipart():
                    for part in message.walk():
                        ctype = part.get_content_type()
                        if ctype == 'image/png':
                            attachment_content = part.get_payload(decode=True)
                            with open('tests/assets/tmp/ysm_mail_attach',
                                      'wb') as fh:
                                fh.write(attachment_content)

                            attachment_correct = filecmp.cmp(
                                'tests/assets/mailcallback_test_file.png',
                                'tests/assets/tmp/ysm_mail_attach')

                        if ctype == 'text/plain':
                            content = bytes.decode(
                                part.get_payload(decode=True))
                            content_correct = (content == content_)

                if content_correct and subject_correct and attachment_correct:
                    message_correct = True
                server.dele(k)
        except Exception:
            raise
        finally:
            server.quit()

        assert message_correct is True

        # ssl security
        subject_ = ''.join(random.choices(string.ascii_lowercase, k=42))
        message_ = ''.join(random.choices(string.ascii_lowercase, k=42))
        test_yaml = f"""
        callbacks:
            mail0:
                type: mail
                host: "{self.SMTP_HOST}"
                port: 465
                login: "{self.SMTP_LOGIN}"
                password: "{self.SMTP_PASSWORD}"
                security: ssl
                from: "{self.SMTP_LOGIN}"
                to: "{self.SMTP_LOGIN}"
                subject: "{subject_}"
                message: "{message_}{{myattr}}"
                attach:
                    - tests/assets/mailcallback_test_file
                delay: 1
        tasks:
            watchfiles:
                type: watchfiles
                changes:
                    - deleted
                paths:
                    - tests/assets/
                callbacks: []
                attrs:
                    myattr: somevalue
        """
        proc = YAMLProcessor()
        proc.load_document(test_yaml)
        callbacks = proc.get_callbacks()
        tasks = proc.get_tasks(callbacks)

        mail = callbacks['mail0']
        task = tasks[0]

        # send out random message
        try:
            loop = asyncio.new_event_loop()
            loop.run_until_complete(mail(task, task.attrs))
            loop.close()
        except Exception:
            self.fail()

        time.sleep(30)  # just to be safe

        # retrive all messages from server and look for our random message
        server = poplib.POP3_SSL(self.POP3_HOST)
        server.user(self.SMTP_LOGIN)
        server.pass_(self.SMTP_PASSWORD)
        content_ = message_ + 'somevalue\n'
        message_correct = False
        try:
            num_messages = len(server.list()[1])
            for k in range(1, num_messages + 1):
                envelope = b'\n'.join(server.retr(k)[1])
                message = email.message_from_bytes(
                        envelope, _class=EmailMessage)

                subject = message['Subject']
                subject_correct = (subject == subject_)

                content_correct = False
                attachment_correct = False
                if message.is_multipart():
                    for part in message.walk():
                        ctype = part.get_content_type()
                        if ctype == 'text/plain':
                            content = part.get_payload(decode=True)
                            if bytes.decode(content) == content_:
                                content_correct = True

                            if attachment_correct is False:
                                with open('tests/assets/tmp/ysm_mail_attach',
                                          'wb') as fh:
                                    fh.write(content)

                                attachment_correct = filecmp.cmp(
                                    'tests/assets/mailcallback_test_file',
                                    'tests/assets/tmp/ysm_mail_attach')
                if content_correct and subject_correct and attachment_correct:
                    message_correct = True
                server.dele(k)
        except Exception:
            raise
        finally:
            server.quit()

        assert message_correct is True

    def test_call_raises_exceptions(self):
        """
        Test that __call__() raises CallbackError on unexpected exceptions.
        """

        test_yaml = f"""
        callbacks:
            mail_stls:
                type: mail
                host: "{self.SMTP_HOST}"
                port: 587
                login: "{self.SMTP_LOGIN}"
                password: "{self.SMTP_PASSWORD}"
                security: starttls
                from: "{self.SMTP_LOGIN}"
                to: "{self.SMTP_LOGIN}"
                subject: "some subject"
                message: "some message"
                attach:
                    - tests/assets/watchfiles_test.sh
            mail_ssl:
                type: mail
                host: "{self.SMTP_HOST}"
                port: 465
                login: "{self.SMTP_LOGIN}"
                password: "{self.SMTP_PASSWORD}"
                security: ssl
                from: "{self.SMTP_LOGIN}"
                to: "{self.SMTP_LOGIN}"
                subject: "some subject"
                message: "some message"
                attach:
                    - tests/assets/watchfiles_test.sh
        tasks:
            watchfiles:
                type: watchfiles
                changes:
                    - deleted
                paths:
                    - tests/assets/
                callbacks: []
                attrs:
                    myattr: somevalue
        """
        proc = YAMLProcessor()
        proc.load_document(test_yaml)
        callbacks = proc.get_callbacks()
        tasks = proc.get_tasks(callbacks)
        task = tasks[0]

        # starttls CallbackError
        mail = callbacks['mail_ssl']
        with mock.patch('smtplib.SMTP_SSL.login') as mck:
            mck.side_effect = Exception
            loop = asyncio.new_event_loop()
            fun = loop.run_until_complete
            with self.assertRaises(CallbackError) as context:
                fun(mail(task, task.attrs))
            err = context.exception
            err_ = "in callback 'mail_ssl' exception 'Exception' was raised, "
            assert str(err) == err_
            loop.close()

        # starttls CallbackError
        mail = callbacks['mail_stls']
        with mock.patch('smtplib.SMTP.starttls') as mck:
            mck.side_effect = Exception
            loop = asyncio.new_event_loop()
            fun = loop.run_until_complete
            with self.assertRaises(CallbackError) as context:
                fun(mail(task, task.attrs))
            err = context.exception
            err_ = "in callback 'mail_stls' exception 'Exception' was raised, "
            assert str(err) == err_
            loop.close()

        # CallbackError
        mail = callbacks['mail_ssl']
        with mock.patch('yasmon.callbacks.process_attributes') as mck:
            mck.side_effect = CallbackError
            loop = asyncio.new_event_loop()
            fun = loop.run_until_complete
            with self.assertRaises(CallbackError) as context:
                fun(mail(task, task.attrs))
            err = context.exception
            err_ = "callback error"
            assert str(err) == err_
            loop.close()

        # CallbackAttributeError
        mail = callbacks['mail_ssl']
        with mock.patch('yasmon.callbacks.process_attributes') as mck:
            mck.side_effect = CallbackAttributeError('attr')
            loop = asyncio.new_event_loop()
            fun = loop.run_until_complete
            with self.assertRaises(CallbackAttributeError) as context:
                fun(mail(task, task.attrs))
            err = context.exception
            err_ = "undefined attribute attr"
            assert str(err) == err_
            loop.close()

        # CallbackCircularAttributeError
        mail = callbacks['mail_ssl']
        with mock.patch('yasmon.callbacks.process_attributes') as mck:
            mck.side_effect = CallbackCircularAttributeError('expr')
            loop = asyncio.new_event_loop()
            fun = loop.run_until_complete
            with self.assertRaises(CallbackCircularAttributeError) as context:
                fun(mail(task, task.attrs))
            err = context.exception
            err_ = "expr\ndetected circular attributes"
            assert str(err) == err_
            loop.close()

        # CallbackError on process_attributes
        mail = callbacks['mail_ssl']
        with mock.patch('builtins.open') as mock_oserror:
            mock_oserror.side_effect = OSError
            loop = asyncio.new_event_loop()
            fun = loop.run_until_complete
            self.assertRaises(CallbackError, fun, mail(task, task.attrs))


if __name__ == '__main__':
    unittest.main(verbosity=2)



from yasmon.callbacks import AbstractCallback
from yasmon.callbacks import CallbackAttributeError
from yasmon.callbacks import CallbackCircularAttributeError

from loguru import logger
from abc import ABC, abstractmethod
from typing import Self, Optional
import watchfiles
import asyncio
import signal
import yaml
import pathlib


class TaskSyntaxError(Exception):
    """
    Raised on task syntax issue.
    """

    def __init__(self, message="task syntax error"):
        self.message = message
        super().__init__(self.message)


class TaskNotImplementedError(Exception):
    """
    Raised if a requested task type is not implemented.
    """

    def __init__(self, type, message="task {type} not implemented"):
        self.message = message.format(type=type)
        super().__init__(self.message)


class AbstractTask(ABC):
    """
    Abstract class from which all task classes are derived.

    Derived tasks are functors calling the assigned callback coroutine
    and can be used for :class:`yasmon.tasks.TaskRunner`.

    The preferred way to instatiate a task is from class
    method :func:`~from_yaml`.
    """
    @abstractmethod
    def __init__(self):
        if not self.name:
            self.name = "Generic Task"
        if not self.attrs:
            self.attrs = {}
        logger.info(f'{self.name} ({self.__class__}) initialized')

    @abstractmethod
    async def __call__(self, callback: AbstractCallback):
        """
        Coroutine called by :class:`TaskRunner`.
        """
        logger.info(f'{self.name} ({self.__class__}) scheduled with '
                    f'{callback.name} ({callback.__class__})')

    @classmethod
    @abstractmethod
    def from_yaml(cls, name: str, data: str,
                  callbacks: list[AbstractCallback]):
        """
        A class method for constructing a callback from a YAML document.

        :param name: unique identifier
        :param data: yaml data
        :param callbacks: collection of callbacks

        :return: new instance
        """
        logger.debug(f'{name} defined from yaml \n{data}')


class TaskList(list):
    """
    A dedicated `list` for tasks.
    """
    def __init__(self, iterable: Optional[list[AbstractTask]] = None):
        if iterable is not None:
            super().__init__(item for item in iterable)
        else:
            super().__init__()


class TaskError(Exception):
    """
    Raised when a path watched by task does not exist anymore.
    """

    def __init__(self, task, message="error in task {task}"):
        self.message = message.format(task=task)
        super().__init__(self.message)


class WatchfilesTask(AbstractTask):
    def __init__(self, name: str, changes: list[watchfiles.Change],
                 callbacks: list[AbstractCallback], paths: list[str],
                 timeout: int, max_retry: int,
                 attrs: Optional[dict[str, str]] = None) -> None:
        """
        :param name: unique identifier
        :param changes: list of watchfiles events
        :param callbacks: assigned callbacks
        :param paths: paths to watch (files/directories)
        :param attrs: (static) attributes
        """
        self.name = name
        self.changes = changes
        self.callbacks = callbacks
        self.paths = paths
        self.abs_paths = []  # resolved upon self.__call__()
        self.attrs = {} if attrs is None else attrs
        self.max_retry = max_retry
        self.timeout = timeout
        super().__init__()

    async def __call__(self, callback):
        await super().__call__(callback)
        retry = 0
        max_retry = 'inf' if self.max_retry == -1 else self.max_retry

        while True:
            try:
                for path in self.paths:
                    if not pathlib.Path.exists(pathlib.Path(path)):
                        raise FileNotFoundError(path)
                    else:
                        abs_path = pathlib.Path(path).resolve()
                        self.abs_paths.append(str(abs_path))
            except FileNotFoundError as path:
                if retry == self.max_retry and self.max_retry > -1:
                    logger.error(f'in task {self.name}'
                                 f' path {path} does not exist '
                                 f'and max_retry {self.max_retry} '
                                 'was reached')
                    raise TaskError(self.name)
                else:
                    retry += 1
                    logger.warning(f'in task {self.name}'
                                   f' path {path} does not exist (anymore)'
                                   f'... retrying callback {callback.name}'
                                   f' after {self.timeout} sec timeout '
                                   f' ({retry}/{max_retry} retries)')
                    await asyncio.sleep(self.timeout)
                    continue

            # run awatch loop if all paths exist
            try:
                await self.awatch_loop(callback)
            except FileNotFoundError:
                continue
            finally:
                retry = 0

    async def awatch_loop(self, callback):
        async for changes in watchfiles.awatch(*self.paths):
            for (change, path) in changes:
                if change in self.changes:
                    match change:
                        case watchfiles.Change.added:
                            chng = 'added'
                        case watchfiles.Change.modified:
                            chng = 'modified'
                        case watchfiles.Change.deleted:
                            chng = 'deleted'
                    try:
                        call_attrs = {'change': chng, 'path': path}
                        await callback(self, self.attrs | call_attrs)
                    except CallbackAttributeError as err:
                        logger.error(f'in task {self.name} callback {callback.name} raised {err}') # noqa
                        raise err
                    except CallbackCircularAttributeError as err:
                        logger.error(f'in task {self.name} callback {callback.name} raised {err}') # noqa
                        raise err

                if (change == watchfiles.Change.deleted and
                        path in self.abs_paths):
                    raise FileNotFoundError

    @classmethod
    def from_yaml(cls, name: str, data: str,
                  callbacks: list[AbstractCallback]) -> Self:
        """
        :class:`WatchfilesTask` can be also constructed from a YAML snippet.

        .. code:: yaml

            changes:
                - added
            callbacks:
                - callback0
            paths:
                - /some/path/to/file1
                - /some/path/to/file2
            attrs:
                myattr: value
            timeout: 10
            max_retry: 3

        Possible changes are ``added``, ``modified`` and ``deleted``.

        :param name: unique identifier
        :param data: YAML snippet
        :param callbacks: list of associated callbacks

        :return: new instance
        :rtype: WatchfilesTask
        """
        super().from_yaml(name, data, callbacks)
        try:
            yamldata = yaml.safe_load(data)
        except yaml.YAMLError as err:
            raise TaskSyntaxError(err)

        if 'timeout' in yamldata:
            try:
                timeout = int(yamldata['timeout'])
                if not timeout > 0:
                    raise ValueError
            except ValueError:
                raise TaskSyntaxError(f"in task {name}: invalid timeout")
            except TypeError:
                raise TaskSyntaxError(f"in task {name}: timeout not integer")
        else:
            timeout = 30

        if 'max_retry' in yamldata:
            try:
                max_retry = int(yamldata['max_retry'])
            except ValueError:
                raise TaskSyntaxError(f"in task {name}: invalid max_retry")
            except TypeError:
                raise TaskSyntaxError(f"in task {name}: max_retry not integer")
        else:
            max_retry = 5

        if 'changes' not in yamldata:
            raise TaskSyntaxError(f"in task {name}: "
                                  "missing changes list")

        if 'paths' not in yamldata:
            raise TaskSyntaxError(f"in task {name}: "
                                  "missing paths list")

        if not isinstance(yamldata['paths'], list):
            raise TaskSyntaxError(f"in task {name}: "
                                  "paths must be a list")

        for path in yamldata['paths']:
            if not isinstance(path, str):
                raise TaskSyntaxError(f"in task {name}: "
                                      "paths must be strings")

        if len(yamldata['paths']) < 1:
            raise TaskSyntaxError(f"in task {name}: "
                                  "at least one path required")

        if 'attrs' in yamldata:
            if not isinstance(yamldata['attrs'], dict):
                raise TaskSyntaxError(f"in task {name}: "
                                      "attrs must be a dictionary")

        imp_changes = ['added', 'modified', 'deleted']
        for change in yamldata['changes']:
            if change not in imp_changes:
                raise TaskSyntaxError(f"in task {name}: "
                                      f"invalid change {change}")

        if not isinstance(yamldata['changes'], list):
            raise TaskSyntaxError(f"in task {name}: "
                                  "changes must be a list")

        changes = [getattr(watchfiles.Change, change)
                   for change in yamldata['changes']]

        paths = yamldata["paths"]
        attrs = yamldata['attrs'] if 'attrs' in yamldata else None
        return cls(name, changes, callbacks, paths, timeout, max_retry, attrs)


class TaskRunner:
    """
    `Asyncio` loop handler. Acts as a functor.
    """
    def __init__(self, tasks: TaskList, testenv: bool = False):
        logger.debug('task runner started...')
        self.loop = asyncio.get_event_loop()
        self.tasks = []

        if testenv:
            self.loop = asyncio.new_event_loop()
            self.tasks.append(
                self.loop.create_task(self.cancle_tasks()))

        signals = (signal.SIGHUP, signal.SIGTERM, signal.SIGINT)
        for s in signals:
            self.loop.add_signal_handler(s, lambda s=s: asyncio.create_task(
                self.signal_handler(s)))

        for task in tasks:
            for callback in task.callbacks:
                self.tasks.append(
                    self.loop.create_task(task(callback)))

    async def __call__(self):
        for task in asyncio.as_completed(self.tasks):
            try:
                await task
            except RuntimeError as err:
                if str(err) == "Already borrowed":
                    # Suppress RuntimeError("Already borrowed"), to
                    # work around this issue:
                    # https://github.com/samuelcolvin/watchfiles/issues/200
                    pass
                else:
                    raise
            except asyncio.CancelledError:
                raise
            except Exception:
                raise

    async def signal_handler(self, sig: signal.Signals):
        """
        Signal handler.
        """
        logger.debug(f'received {sig.name}')
        for task in self.tasks:
            task.cancel()

    async def cancle_tasks(self, delay=2):
        """
        Cancle tasks (for testing purposes)
        """
        await asyncio.sleep(delay)
        for task in self.tasks:
            if task is not asyncio.current_task():
                task.cancel()



from yasmon.callbacks import AbstractCallback
from yasmon.callbacks import CallbackDict
from yasmon.callbacks import ShellCallback, LoggerCallback, MailCallback
from yasmon.callbacks import CallbackSyntaxError
from yasmon.callbacks import CallbackNotImplementedError
from yasmon.tasks import TaskList, WatchfilesTask, TaskNotImplementedError
from yasmon.loggers import LoggerSyntaxError, AbstractLogger
from yasmon.loggers import StdErrLogger, FileLogger, JournalLogger

from loguru import logger
import yaml


class YAMLSyntaxError(Exception):
    """
    Raised when on general YAML syntax errors.
    """

    def __init__(self, message="yaml syntax error"):
        self.message = message
        super().__init__(self.message)


class YAMLProcessor:
    def __init__(self):
        self.data = None

    def load_file(self, filename: str):
        try:
            fh = open(filename, "r")
            logger.info(f'using config file {filename}')
        except FileNotFoundError as err:
            logger.error(f"YAML file {filename} not found")
            raise err
        except PermissionError as err:
            logger.error(
                f"{err.__class__.__name__} while opening {filename} {err}")
            raise err
        except OSError as err:
            logger.error(
                f"{err.__class__.__name__} while opening {filename} {err}")
            raise err
        except Exception as err:
            logger.error(
                f"{err.__class__.__name__} while opening {filename} {err}")
            raise err
        else:
            try:
                self.data = yaml.safe_load(fh)
            except yaml.YAMLError as err:
                raise YAMLSyntaxError(err)
            finally:
                fh.close()

            if self.data is None:
                raise YAMLSyntaxError('config file is empty')

    def add_loggers(self) -> list[AbstractLogger]:
        """
        Add defined loggers.

        :return: list of added loggers or empty list
        :rtype: list[AbstractLogger]
        """
        logger.info('processing loggers...')
        loggers: list[AbstractLogger] = []

        for key in self.data:
            data = self.data[key]
            yaml_data = yaml.dump(data)
            try:
                match key:
                    case 'log_stderr':
                        instance = StdErrLogger.from_yaml(yaml_data)
                        loggers.append(instance)
                    case 'log_file':
                        instance = FileLogger.from_yaml(yaml_data)
                        loggers.append(instance)
                    case 'log_journal':
                        instance = JournalLogger.from_yaml(yaml_data)
                        loggers.append(instance)
            except LoggerSyntaxError:
                raise

        return loggers

    def load_document(self, document: str):
        try:
            self.data = yaml.safe_load(document)
            logger.info(f'config:\n{document}')
        except yaml.YAMLError as err:
            raise YAMLSyntaxError(err)

        if self.data is None:
            raise YAMLSyntaxError('config document is empty')

    def get_tasks(self, callbacks: CallbackDict):
        logger.debug('processing tasks...')
        if 'tasks' not in self.data:
            raise YAMLSyntaxError('tasks not defined')

        if not self.data['tasks']:
            logger.warning('no tasks defined')
            return TaskList()

        tasks = self.data['tasks']

        if type(tasks) is not dict:
            raise YAMLSyntaxError('tasks must be a dictionary')

        taskslist = TaskList()
        for task in tasks:
            taskdata = tasks[task]
            if type(taskdata) is not dict:
                raise YAMLSyntaxError(f'{task} task data must be a dictionary')

            taskdata_yaml = yaml.dump(taskdata)

            if 'callbacks' not in taskdata:
                raise YAMLSyntaxError(f'{task} task data must include'
                                      ' callbacks list')

            task_callbacks: list[AbstractCallback] = [
                callbacks[c] for c in taskdata["callbacks"]
                if c in taskdata["callbacks"]
            ]

            match taskdata['type']:
                case 'watchfiles':
                    taskslist.append(WatchfilesTask.from_yaml(task,
                                                              taskdata_yaml,
                                                              task_callbacks))
                case _:
                    raise TaskNotImplementedError(
                        f'task type {taskdata["type"]}'
                        ' not implement')

        logger.debug('done processing tasks')
        return taskslist

    def get_callbacks(self):
        logger.debug('processing callbacks...')
        if 'callbacks' not in self.data:
            raise YAMLSyntaxError('callbacks not defined')

        if not self.data['callbacks']:
            logger.warning('no callbacks defined')
            return

        callbacks = self.data['callbacks']

        if type(self.data['callbacks']) is not dict:
            raise YAMLSyntaxError('callbacks must be a dictionary')

        callbacksdict = CallbackDict()
        for callback in callbacks:
            callbackdata = self.data['callbacks'].get(callback)
            if type(callbackdata) is not dict:
                raise YAMLSyntaxError(f'{callback} callback data must'
                                      ' be a dictionary')

            try:
                match callbackdata['type']:
                    case 'shell':
                        callbacksdict[callback] = ShellCallback.from_yaml(
                            callback, yaml.dump(callbackdata))
                    case 'logger':
                        callbacksdict[callback] = LoggerCallback.from_yaml(
                            callback, yaml.dump(callbackdata))
                    case 'mail':
                        callbacksdict[callback] = MailCallback.from_yaml(
                            callback, yaml.dump(callbackdata))
                    case _:
                        raise CallbackNotImplementedError(
                            f'callback type {callbackdata["type"]} '
                            'not implement')
            except CallbackSyntaxError as err:
                logger.error(f'error while processing callbacks: {err}. '
                             'Exiting!')
                raise err

        logger.debug('done processing callbacks')
        return callbacksdict
